# -*- coding: utf-8 -*

"""
A template of configuration file for TNT.
"""

# 1. Blocks to be added.
new_blocks = set(['NAMNEW',
                  ])
# 2. Blocks to be moved. If target block exists, raise an error.
blocks_to_move = {'NAMOLD': 'NAMMOVED',
                  }
# 3. Keys to be moved. If target exists or target block is missing, raise an error.
# Blocks need to be consistent with above blocks movings.
keys_to_move = {('NAMMOVED', 'KEYOLD'): ('NAMNEW', 'KEYNEW'),  # change the key from block, and/or rename it
                }
# 4. Keys to be removed. Already missing keys are ignored.
# Blocks need to be consistent with above movings.
keys_to_remove = set([('NAMBLOCK', 'KEYTOREMOVE'),
                      ])
# 5. Keys to be set with a value (new or modified). If block is missing, raise an error.
# Blocks need to be consistent with above movings.
keys_to_set = {('NAMBLOCK1', 'KEY1'): 46.5,
               ('NAMBLOCK2', 'KEY2(1:3)'): [5, 6, 7],
               ('NAMBLOCK3', 'KEY3(50)'): -50,
               }
# 6. Blocks to be removed. Already missing blocks are ignored.
blocks_to_remove = set(['NAMBLOCK',
                        ])
# 7. Macros: substitutions in the namelist's values. A *None* value ignores
# the substitution (keeps the keyword, to be substituted later on).
macros = {'VAL_TO_SUBSTITUTE': 8,
          'VAL_TO_KEEP_AND_BE_SUBSTITUTED_LATER': None}



"""
Configuration file for the Sphinx documentation builder.
"""

#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
import os
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))

# This will make sphinx crash if something goes wrong with bronx import !
import thenamelisttool

assert thenamelisttool

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'thenamelisttool'
copyright = '2022, The Vortex Team'
author = 'The Vortex Team'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',       # Core library for html generation from docstrings
    'sphinx.ext.autosummary',   # Create neat summary tables
    'sphinx.ext.viewcode',
]
autosummary_generate = True  # Turn on sphinx.ext.autosummary

# Concatenate the class docstring and the __init__ docstring
autoclass_content = 'both'

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinxdoc'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
TNTdiff - The Namelist Tool: a namelist comparator.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import os
import sys

# Automatically set the python path
sys.path.insert(
    0,
    os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'src')
)

from thenamelisttool.entrypoints import tntdiff as tnt_cli


if __name__ == '__main__':
    tnt_cli.main()



# -*- coding: utf-8 -*-

"""
TNTdiffpack - The Namelist Tool: a namelist pack comparator.

Compares two namelists packs and produces a summary on the standard output
and in a separate file.

Beware that TNTdiffpack (purposely) DO NOT take into account:

* Differences in the order of appearance of namelists blocks or keys
* Differences in the formatting of namelist values (e.g 1.0 and 1.0000 are
  considered the same)

The displayed namelists DO NOT necessarily correspond to the original files since,
prior to be displayed, blocks/keys are ordered alphabetically and values are
formatted in a "standard" way.

"""

from __future__ import print_function, absolute_import, unicode_literals, division

import argparse
import difflib
import io
import os
import sys


from bronx.fancies.display import printstatus
from bronx.stdtypes.tracking import MappingTracker
import thenamelisttool as tnt

_outfilename = 'tntdiffpack.out'


def _compute_diffs(nambefore, namafter, modified):
    diffs = dict()
    for k in modified:
        txtB = nambefore[k].split('\n')
        txtA = namafter[k].split('\n')
        diffs[k] = difflib.ndiff(txtB, txtA)
    return diffs


def main():
    """Run the tntdiffpack CLI."""
    program_desc = '%(prog)s -- ' + __import__('__main__').__doc__.lstrip('\n')
    parser = argparse.ArgumentParser(description=program_desc, epilog='End of help for: %(prog)s',
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-b', '--before',
                        required=True,
                        help="source namelist pack.")
    parser.add_argument('-a', '--after',
                        required=True,
                        help="target namelist pack.")
    parser.add_argument('-o',
                        default=_outfilename,
                        dest='outputfilename',
                        help="output filename (without any extension). Defaults to %(default)s.")
    args = parser.parse_args()

    ko = set()
    nambefore = dict()
    namafter = dict()
    listbefore = [f for f in os.listdir(args.before)
                  if os.path.isfile(os.path.join(args.before, f))]
    listafter = [f for f in os.listdir(args.after)
                 if os.path.isfile(os.path.join(args.after, f))]
    listcommon = set(listbefore) & set(listafter)

    for (targetdict, listdir, inputdir) in ((nambefore, listbefore, args.before),
                                            (namafter, listafter, args.after)):
        sys.stdout.write('Processing files in {:s}: '.format(inputdir))
        for i, f in enumerate(listdir):
            printstatus(i + 1, len(listdir))
            if f in listcommon:
                try:
                    nparsed = tnt.util.namelist_read_and_sort(os.path.join(inputdir, f))
                except ValueError:
                    ko.add(f)
                else:
                    targetdict[f] = nparsed
            else:
                targetdict[f] = ''

    tracker = MappingTracker(nambefore, namafter)
    computediffs = _compute_diffs(nambefore, namafter, tracker.updated)
    # Expand the generator objects into lists
    print('Creating diff outputs. It may take a while (depending on the amount of changes).')
    for i, n in enumerate(tracker.updated):
        computediffs[n] = [l for l in computediffs[n]]

    outtpl = tnt.config.get_template('tnt-diffpack-output.tpl', encoding='utf_8')

    with io.open(args.outputfilename, "w") as fhout:
        fhout.write(outtpl.substitute(ref=args.before, new=args.after,
                                      ko='\n'.join(['{:s}'.format(n) for n in sorted(ko)]),
                                      untouched='\n'.join(['{:s}'.format(n) for n in sorted(tracker.unchanged)]),
                                      created='\n'.join(['{:s}'.format(n) for n in sorted(tracker.created)]),
                                      deleted='\n'.join(['{:s}'.format(n) for n in sorted(tracker.deleted)]),
                                      modified='\n'.join(['{:s}'.format(n) for n in sorted(tracker.updated)]),
                                      computediffs='\n'.join(['============ {:s} ============\n\n{:s}\n'.
                                                              format(n, '\n'.join(computediffs[n]))
                                                              for n in tracker.updated])
                                      ))



from __future__ import print_function, absolute_import, unicode_literals, division

import os
import unittest

from bronx.fancies import loggers

from thenamelisttool.config import TntDirective, TntStackDirective
from thenamelisttool.config import TntDirectiveUnkownError, TntDirectiveValueError, TntStackDirectiveError

tpl_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        '../src/thenamelisttool/templates')
tpl_path = os.path.normpath(tpl_path)


class TestTntConfig(unittest.TestCase):

    def test_tnt_dir(self):
        with self.assertRaises(TntDirectiveUnkownError):
            TntDirective(scrontch='toto')
        with self.assertRaises(TntDirectiveUnkownError):
            TntDirective().hips
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(keys_to_set='I believe, I can fly')
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(keys_to_set=dict(toto=1))
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(keys_to_set={(25, 42): 1})
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(keys_to_set=dict(BLOP={1: 25}))
        for tdir in [TntDirective(keys_to_set={('BLOP', 'TRUC'): 1}),
                     TntDirective(keys_to_set=dict(BLOP={'TRUC': 1}))]:
            self.assertDictEqual(tdir.keys_to_set, {('BLOP', 'TRUC'): 1})
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(blocks_to_move=1)
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(blocks_to_move={'ARG': 1})
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(blocks_to_move={1: 'ARG'})
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(blocks_to_remove=1)
        self.assertSetEqual(TntDirective(blocks_to_remove='ARG').blocks_to_remove,
                            set(['ARG', ]))
        self.assertSetEqual(TntDirective(blocks_to_remove=('ARG', 'BLOP')).blocks_to_remove,
                            set(['ARG', 'BLOP']))
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(macros='MACHIN')
        with self.assertRaises(TntDirectiveValueError):
            TntDirective(macros={1: 3})

    def test_tntstack_dir(self):
        with self.assertRaises(TntStackDirectiveError):
            TntStackDirective(tpl_path, todolist='toto')
        with self.assertRaises(TntStackDirectiveError):
            TntStackDirective(tpl_path, list(), directives='Beeeeee')
        with self.assertRaises(TntStackDirectiveError):
            TntStackDirective(tpl_path, [dict(action='truc'), ])
        with self.assertRaises(TntStackDirectiveError):
            TntStackDirective(tpl_path, [dict(namelist='machin'), ])
        with loggers.contextboundGlobalLevel('critical'):
            with self.assertRaises(TntStackDirectiveError):
                # Namelist is missing
                TntStackDirective(tpl_path, [dict(action='tnt', directive='truc'), ])
        with self.assertRaises(TntStackDirectiveError):
            TntStackDirective(tpl_path, [dict(action='tnt', namelist=1, directive='truc'), ])
        with self.assertRaises(TntStackDirectiveError):
            # machin directive is not defined
            TntStackDirective(tpl_path,
                              [dict(action='tnt', namelist='coucou', directive='machin'), ],
                              directives=dict(truc=dict()))
        tdir = TntStackDirective(tpl_path,
                                 [dict(action='tnt', namelist='coucou', directive='truc'), ],
                                 directives=dict(truc=dict()))
        self.assertDictEqual(tdir.todolist[0],
                             dict(action='tnt', namelist=['coucou', ], directive=['truc', ]))
        with loggers.contextboundGlobalLevel('critical'):
            with self.assertRaises(TntStackDirectiveError):
                # multiple target...
                TntStackDirective(tpl_path,
                                  [dict(action='create', target=['blop', 'toto', ],
                                        external='namelist_prep_template'), ])
        with self.assertRaises(TntStackDirectiveError):
            # namelist_prep_template2 do not exists
            TntStackDirective(tpl_path,
                              [dict(action='create', target='blop', external='namelist_prep_template2'), ])
        tdir = TntStackDirective(tpl_path,
                                 [dict(action='create', target='blop', external='namelist_prep_template'), ])
        self.assertDictEqual(tdir.todolist[0],
                             dict(action='create', target='blop',
                                  external=os.path.join(tpl_path, 'namelist_prep_template')))


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

"""
TNT Stack - A namelist's pack updater.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import argparse
import io
import os
import yaml

import thenamelisttool as tnt
from thenamelisttool.namadapter import NO_SORTING, FIRST_ORDER_SORTING, SECOND_ORDER_SORTING

_tmpl = 'tmpl_directives.tntstack.yaml'


def main():
    """Run the tntstack CLI."""
    parser = argparse.ArgumentParser(description="TNT Stack - A namelist's pack updater.",
                                     epilog='End of help for: %(prog)s')

    parser.add_argument('-v',
                        action='store_true',
                        dest='verbose',
                        help='verbose mode.',
                        default=False)
    sorting = parser.add_mutually_exclusive_group()
    sorting.add_argument('-S',
                         action='store_const', const=FIRST_ORDER_SORTING + 1,
                         dest='first_order_sorting',
                         help='first order sorting: sort all keys within blocks.')
    sorting.add_argument('-0',
                         action='store_const', const=NO_SORTING + 1,
                         dest='no_sorting',
                         help='no sorting at all (the default is second_order_sorting, \
                               i.e. sort only within indexes or attributes of the \
                               same key within blocks).')
    directive = parser.add_mutually_exclusive_group(required=True)
    directive.add_argument('-d',
                           dest='directive',
                           type=str,
                           help='the file in which update directives are stored. \
                                 Activate option -D instead of -d to generate a template.')
    directive.add_argument('-D',
                           dest='generate_directive_template',
                           action='store_true',
                           help="generates a directive template written in '{}'.".format(_tmpl))
    args = parser.parse_args()

    if args.generate_directive_template:
        tnt.config.write_directives_template(_tmpl, tplname='tntstack-directive.tpl.yaml')
        print("Template of directives written in: " + os.path.abspath(_tmpl))
    else:
        # Find the basedir
        dirpath = os.path.realpath(args.directive)
        basedir = os.path.dirname(dirpath)

        with io.open(args.directive, 'r') as fhyaml:
            directive = tnt.config.TntStackDirective(basedir, ** yaml.load(fhyaml))

        with tnt.util.set_verbose(args.verbose, args.directive):
            tnt.util.process_tnt_stack(directive,
                                       sorting=(args.first_order_sorting or args.no_sorting or
                                                SECOND_ORDER_SORTING + 1) - 1)



# -*- coding: utf-8 -*-

"""
TNT - The Namelist Tool: a namelist updater.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import argparse
import io
import os

import thenamelisttool as tnt


def main():
    """Start the tnt CLI."""
    _tmpl = 'tmpl_directives.tnt'

    parser = argparse.ArgumentParser(description='TNT - The Namelist Tool: a namelist updater.',
                                     epilog='End of help for: %(prog)s')
    parser.add_argument('namelists',
                        type=str,
                        nargs='*',
                        help='namelist(s) file(s) to be processed.')
    directives = parser.add_mutually_exclusive_group(required=True)
    directives.add_argument('-d',
                            dest='directives',
                            type=str,
                            help='the file in which update directives are stored. \
                                  Activate option -D instead of -d to generate a template.')
    directives.add_argument('-D',
                            dest='generate_directives_template',
                            action='store_true',
                            help="generates a directives template '{}.(py|yaml)'.".format(_tmpl))
    directives.add_argument('-n',
                            dest='namdelta',
                            type=str,
                            help='path to a file that contains a namelist delta.')
    directives.add_argument('-c',
                            dest='check_namelist',
                            action='store_true',
                            help='check that the namelist is ok and do a first order sorting. \
                                  This option is equivalent to. \
                                  "tnt.py -d void.py -S NAMELIST" with an empty void.py directive file.',
                            default=False)
    directives.add_argument('--squeeze',
                            dest='squeeze',
                            action='store_true',
                            help='squeeze the namelist: remove empty blocks.',
                            default=False)
    parser.add_argument('-i',
                        action='store_true',
                        dest='in_place',
                        help='modifies the namelist(s) in place. \
                              Else, modified namelists are suffixed with .tnt \
                              if not given as arg -o.',
                        default=False)
    parser.add_argument('-o',
                        dest='outfilename',
                        default=None,
                        help="directives (.py) output filename.")
    sorting = parser.add_mutually_exclusive_group()
    sorting.add_argument('-S',
                         action='store_true',
                         dest='firstorder_sorting',
                         help='First order sorting: sort all keys within blocks.',
                         default=False)
    sorting.add_argument('-s',
                         action='store_true',
                         dest='secondorder_sorting',
                         help='Second order sorting: sort only within indexes \
                               or attributes of the same key within blocks.',
                         default=False)
    parser.add_argument('--doctor',
                        action='store_true',
                        dest='doctor',
                        help='try to convert value to DOCTOR norm according \
                              type for moved keys',
                        default=False)
    parser.add_argument('--keep_index',
                        action='store_true',
                        dest='keep_index',
                        help='moved keys in identical block keep the original \
                              index of key in block (except a sorting is \
                              requested)',
                        default=False)
    parser.add_argument('-r',
                        dest='blocks_ref',
                        type=str,
                        help='a (possibly empty) reference namelist, to which the \
                              set of blocks is asserted to be equal.',
                        default=None)
    parser.add_argument('-v',
                        action='store_true',
                        dest='verbose',
                        help='verbose mode.',
                        default=False)
    args = parser.parse_args()

    if args.firstorder_sorting or args.check_namelist:
        sorting = tnt.namadapter.FIRST_ORDER_SORTING
    elif args.secondorder_sorting:
        sorting = tnt.namadapter.SECOND_ORDER_SORTING
    else:
        sorting = tnt.namadapter.NO_SORTING

    if len(args.namelists) > 1 and args.outfilename is not None:
        raise ValueError('Arg -o should not be used applied to several namelists')

    if args.generate_directives_template:
        tnt.config.write_directives_template(_tmpl + '.py', tplname='tnt-directive.tpl.py')
        print("Template of directives written in: " + os.path.abspath(_tmpl + '.py'))
        tnt.config.write_directives_template(_tmpl + '.yaml', tplname='tnt-directive.tpl.yaml')
        print("Template of directives written in: " + os.path.abspath(_tmpl + '.yaml'))
    else:
        assert len(args.namelists) > 0, "no namelists provided to process."
        if args.directives:
            directives = tnt.config.read_directives(args.directives)
        else:
            if args.check_namelist or args.squeeze:
                directives = tnt.config.TntDirective()
            else:
                with io.open(args.namdelta, 'r') as fhnam:
                    directives = tnt.config.TntDirective(namdelta=fhnam.read())
        for nam in args.namelists:
            with tnt.util.set_verbose(args.verbose, nam):
                tnt.util.process_namelist(nam, directives,
                                          sorting=sorting,
                                          in_place=args.in_place,
                                          outfilename=args.outfilename,
                                          blocks_ref=args.blocks_ref,
                                          doctor=args.doctor,
                                          keep_index=args.keep_index,
                                          squeeze=args.squeeze)



from __future__ import print_function, absolute_import, unicode_literals, division

import io
import os
import unittest

import thenamelisttool as tnt

tloglevel = 9999

tpl_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        '../src/thenamelisttool/templates')
tpl_path = os.path.normpath(tpl_path)

data_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
data_path = os.path.normpath(data_path)

tntstack_todo_ref = [
    {'action': 'tnt', u'namelist': [u'namelist_screen*'], u'directive': [u'dfi', u'geo499c1']},
    {'action': 'tnt', u'namelist': [u'namelist_previ_sfx', u'namelist_surf'], u'directive': [u'surfexdiags']},
    {'action': 'create', u'copy': u'namelist_screening1', u'target': u'namelist_screening3'},
    {'action': 'create', u'namelist': u'namelist_fp1', u'target': u'namelist_fp2', u'directive': [u'geo499c1', u'surfexdiags']},
    {'action': 'create', u'target': u'namelist_prep', u'external': os.path.join(tpl_path, u'namelist_prep_template')},
    {'action': 'delete', u'namelist': [u'something_useless[12]', u'something_strange']},
    {'action': 'link', u'namelist': u'namelist_prep', u'target': u'namelist_fp3'},
    {'action': 'move', u'namelist': u'namelist_surf', u'target': u'namelist_surfex'},
    {'action': 'touch', u'namelist': [u'unknown_namelist', u'namelist_fp*']},
    {'action': 'clean_untouched'}]

COMPOSED_NAM = """\
 &NAERAD
 /
 &NAMCT0
   LECMWF=.FALSE.,
 /
 &NAMDIM
   NPROMA=-24,
 /
 &NAMDYN
   LNHDYN=.TRUE.,
 /
 &NAMEMPTY
 /
 &NAMFA
   NBITCS=16,
   NBITPG=24,
   NVGRIB=123,
 /
 &NAMINI
   NSTEP=8,
 /
 &NAMKEPT
   USELESS='YES',
 /
 &NAMOBJ
   L_OOPS=.TRUE.,
 /
 &NAMOBS
   LOLDPP=.TRUE.,
 /
 &NAMRIP
   CSTOP='h1',
 /
"""


def checklib_yaml():
    rc = True
    try:
        import yaml  # @UnusedImport
    except ImportError:
        rc = False
    return rc


class TestTntTemplate(unittest.TestCase):

    def test_tnt_tpl_py(self):
        tplpy = tnt.config.read_directives(os.path.join(tpl_path,
                                                        'tnt-directive.tpl.py'))
        self.assertDictEqual(tplpy.keys_to_set,
                             {('NAMBLOCK2', 'KEY2(1:3)'): [5, 6, 7],
                              ('NAMBLOCK1', 'KEY1'): 46.5,
                              ('NAMBLOCK3', 'KEY3(50)'): -50})

    @unittest.skipUnless(checklib_yaml(), "pyyaml is unavailable")
    def test_tnt_tpl_yaml(self):
        tplyaml = tnt.config.read_directives(os.path.join(tpl_path,
                                                          'tnt-directive.tpl.yaml'))
        self.assertDictEqual(tplyaml.keys_to_set,
                             {('NAMBLOCK2', 'KEY2(1:3)'): [5, 6, 7],
                              ('NAMBLOCK3', 'KEY3(50)'): -50,
                              ('NAMBLOCK1', 'KEY1'): 46.5})

    @unittest.skipUnless(checklib_yaml(), "pyyaml is unavailable")
    def test_tntstack_tpl_yaml(self):
        # Read the yaml file
        import yaml
        with io.open(os.path.join(tpl_path, 'tntstack-directive.tpl.yaml')) as fhy:
            dirdict = yaml.load(fhy, Loader=yaml.SafeLoader)
        # Process it
        tplyaml = tnt.config.TntStackDirective(basedir=tpl_path, **dirdict)
        self.assertListEqual(tplyaml.todolist, tntstack_todo_ref)
        self.assertSetEqual(set(tplyaml.directives.keys()),
                            set(['surfexdiags', 'geo499c1', 'dfi']))


@unittest.skipUnless(checklib_yaml(), "pyyaml is unavailable")
class TestTntRecipe(unittest.TestCase):

    def test_recipe_yaml(self):
        from bronx.datagrip.namelist import FIRST_ORDER_SORTING
        recipe = tnt.config.TntRecipe(os.path.join(tpl_path, 'tntcompose-recipe.tpl.yaml'),
                                      sourcenam_directory=data_path)
        nam = recipe.ingredients[0]
        for ingredient in recipe.ingredients[1:]:
            nam.merge(ingredient)
        self.assertEqual(nam.dumps(sorting=FIRST_ORDER_SORTING), COMPOSED_NAM)

    def _assert_syntax_error(self, recipe):
        with tnt.util.set_verbose(False, 'ko/recipe_ko1.yaml'):
            with self.assertRaises(tnt.config.TntRecipeSyntaxError):
                tnt.config.TntRecipe(os.path.join(data_path, 'ko', recipe),
                                     sourcenam_directory=data_path)

    def test_recipe_syntax(self):
        self._assert_syntax_error('recipe_ko1.yaml')
        self._assert_syntax_error('recipe_ko2.yaml')
        self._assert_syntax_error('recipe_ko3.yaml')
        self._assert_syntax_error('recipe_ko5.yaml')
        self._assert_syntax_error('recipe_ko6.yaml')
        self._assert_syntax_error('recipe_ko7.yaml')


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

"""
TNT stands for "The Namelist Tool".

The main objectives of this package is to provide command line tools useful to
work with namelists or pack of namelists (update them, compare them, ...).

The Sphinx documentation of this package is few. However, for end users, the
entry points to this package are the ``tnt.py``, ``tntdiff.py`` and
``tntstack.py`` command line utilities provided in the ``tools`` subdirectory
(or in your ``$PATH`` when ``pip`` is used).
These command line utility are provided with an embedded documentation accessible
with the `-h` option (e.g. ``tnt.py -h``).

The ``tnt.py`` and ``tntstack.py`` command line utilities heavily rely on
directive files. A template of such directive files can be generated using the
``-D`` option (``tnt.py -D`` or ``tntstack.py -D``). The template directives
files are thoroughly documented and should be regarded as documentation.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

from . import config
from . import namadapter
from . import util



# -*- coding: utf-8 -*-

"""
Adapter classes for namelist's parsers.

Set of adapter classes that uses external namelist's parsers to provide
functionalities that fit the needs of TNT utilities.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import abc
import re

import six

from bronx.compat.moves import collections_abc
from bronx.fancies import loggers

tntlog = loggers.getLogger('tntlog')

# Macros List from vortex's common.data.namelists
#: TNT's predefined list of macros
KNOWN_NAMELIST_MACROS = set(['NPROC', 'NBPROC', 'NBPROC_IO', 'NCPROC', 'NDPROC',
                             'NBPROCIN', 'NBPROCOUT', 'IDAT', 'CEXP',
                             'TIMESTEP', 'FCSTOP', 'NMODVAL', 'NBE', 'SEED',
                             'MEMBER', 'NUMOD', 'OUTPUTID', 'NRESX', 'PERTURB',
                             'JOUR', 'RES', 'LLADAJ', 'LLADMON', 'LLFLAG',
                             'LLARO', 'LLVRP', 'LLCAN', ])
# Other known namelist's macros (not to be substituted)
KNOWN_NAMELIST_MACROS.update(['substr6', 'substrA', 'substrC', 'XMP_TYPE',
                              'XLOPT_SCALAR', 'XNCOMBFLEN', 'val_sitr', 'val_sipr',
                              '_lbias_', '_lincr_'])

#: Output namelist sorting option: NO_SORTING. Keep the initial ordering.
NO_SORTING = 0
#: Output namelist sorting option: FIRST_ORDER_SORTING. Sort all keys within blocks.
FIRST_ORDER_SORTING = 1
#: Output namelist sorting option: SECOND_ORDER_SORTING. Sort only between indexes.
SECOND_ORDER_SORTING = 2


@six.add_metaclass(abc.ABCMeta)
class AbstractNamelistAdapter(collections_abc.Mapping):
    """Every Namelist adapter must derive from this abstract class."""

    @abc.abstractmethod
    def __init__(self, namelistsfile, macros=None):  # @UnusedVariable
        """
        :param str namelistsfile: The namelist itself or a path to a namelist file.
        """
        self._parser = None

    @property
    def parser(self):
        """The internal namelist's parser."""
        return self._parser

    # Public methods that operates on namelist's blocks

    def add_blocks(self, blocks):
        """Add a set of new blocks in the present namelist's set.

        :param list[str] blocks: ['BLOCK1', 'BLOCK2', ...]
        """
        for b in blocks:
            if b not in self:
                self._actual_newblock(b)
            else:
                tntlog.info('block "%s" is already present.', b)

    def remove_blocks(self, blocks):
        """Remove a set of blocks from the present namelist's set.

        :param list[str] blocks: ['BLOCK1', 'BLOCK2', ...]
        """
        for b in blocks:
            if b in self:
                self._actual_rmblock(b)
            else:
                tntlog.info('block "%s" to be removed but already missing.', b)

    def move_blocks(self, blocks):
        """Move/Rename a set of blocks within the present namelist's set.

        :param dict[str] blocks: {'BLOCK_OLD':'BLOCK_NEW', ...}
        """
        for (old_b, new_b) in blocks.items():
            if old_b in self:
                if new_b not in self:
                    self._actual_mvblock(old_b, new_b)
                else:
                    raise ValueError('block "{:s} already present'.format(new_b))
            else:
                tntlog.warning('block "%s" to be moved but missing from namelist: ignored.',
                               old_b)

    def check_blocks(self, another, macros=None):
        """
        Check that the present namelist's set contains the same set of blocks as
        **another** does.

        **another** can be either the filename of a namelist to be read, or any
        kind of :class:`AbstractNamelistAdapter` instance.

        If **macros** is not None, it can contain the macros a.k.a. values to be
        replaced, e.g.: {'NPROC':8, 'substrA':None} will replace all NPROC values
        by 8 and will let substrA untouched.

        :return set: The set of blocks that differ.
        """
        if not isinstance(another, AbstractNamelistAdapter):
            another = self.__class__(another, macros=macros)
        return set(self.keys()).symmetric_difference(set(another.keys()))

    # Public methods that operates on namelist's keys

    def add_keys(self, keys, doctor=False, indexes=None):
        """Set a set of keys in the present namelist's set.

        :param dict[tuple] keys: {('BLOCK','KEY'):value, ...}
        :param bool doctor: if true, try to convert value to DOCTOR norm according type
        :param dict[tuple] indexes: if present, set the keys at given index in block
                                    {('BLOCK','KEY'):index, ...}
        """
        if indexes is None:
            indexes = {}
        for ((b, k), v) in keys.items():
            idx = indexes.get((b, k), None)
            if b in self:
                self._actual_newkey(b, k,
                                    self._DOCTOR_convert(k, v) if doctor else v,
                                    index=idx)
            else:
                raise KeyError('block "{:s}" is missing: cannot set its "{:s}" key'
                               .format(b, k))

    def remove_keys(self, keys):
        """Remove a set of keys from the present namelist's set.

        :param list[tuple] keys: [('BLOCK1','KEY1'), ('BLOCK2','KEY2'), ...]
        """
        for (b, k) in self._expand_keys(keys):
            if b in self:
                if k in self[b]:
                    self._actual_rmkey(b, k)
                else:
                    tntlog.info(('key "%s" to be removed but already missing from block "%s".',
                                 k, b))
            else:
                tntlog.info('block "%s" missing: cannot remove its "%s" key.',
                            b, k)

    def move_keys(self, keys, doctor=False, keep_index=False):
        """
        Move a set of keys within the present namelist's set.

        :param dict[tuple] keys: {('BLOCK_OLD','KEY_OLD'):('BLOCK_NEW','KEY_NEW'), ...}
        :param bool doctor: if True, try to convert value to DOCTOR norm according type
        :param bool keep_index: if True, moved keys in identical block keep the
                                original index of key in block (except a sorting
                                is requested later on.
        """
        origin_keys = self._expand_keys(keys.keys(), radics=True)
        expanded_keys = {}
        for (ob, o_r, ok) in origin_keys:
            (nb, n_r) = keys[(ob, o_r)]
            expanded_keys[(ob, ok)] = (nb, ok.replace(o_r, n_r, 1))
        for (ob, ok), (nb, nk) in expanded_keys.items():
            if ob in self:
                if ok in self[ob]:
                    if keep_index and ob == nb:
                        idx = list(self[ob].keys()).index(ok)
                    else:
                        idx = None
                    v = self[ob][ok]
                    self.remove_keys([(ob, ok)])
                    if nb in self:
                        if nk not in self[nb]:
                            self.add_keys({(nb, nk): v},
                                          doctor=doctor, indexes={(nb, nk): idx})
                        else:
                            raise ValueError(('key "{:s}" in block "{:s}" ' +
                                              'already exists: prevent moving from ' +
                                              'block "{:s}" key "{:s}".')
                                             .format(nk, nb, ob, ok))
                    else:
                        raise KeyError(('block "{:s}" missing: cannot move key "{:s}"' +
                                        'from block "{:s}" to it as key "{:s}.')
                                       .format(nb, ok, ob, nk))
                else:
                    tntlog.warning('key "%s" missing from block "%s": cannot move it.',
                                   ok, ob)
            else:
                tntlog.warning('block "%s" missing: cannot move its key "%s".', ob, ok)

    def squeeze(self):
        """Squeeze the namelist: remove empty blocks."""
        self._actual_squeeze()

    # Generic utility methods

    @staticmethod
    def _all_macros(arg_macros):
        macros = {k: None for k in KNOWN_NAMELIST_MACROS}
        if arg_macros is not None:
            macros.update(arg_macros)
        return macros

    def _expand_keys(self, keys, radics=False):
        """
        Find all entries corresponding to the given keys,
        due to attributes and/or indexes.

        :param list[tuple] keys: [('BLOCK1','KEY1'), ('BLOCK2','KEY2'), ...]
        :param bool radics: add the radical in the tuples
        :return list[tuple]: The expanded list of namelist keys
        """
        expanded_keys = []
        for (b, k) in keys:
            if b in self:
                ek = [(b, nk) for nk in self[b].keys()
                      if re.match(k.replace('(', r'\(').replace(')', r'\)') + r'(\(.+\)|%.+)*$', nk)]
                if radics:
                    ek = [(b, k, nk) for (b, nk) in ek]
                expanded_keys.extend(ek)
        return set(expanded_keys)

    @staticmethod
    def _DOCTOR_convert(key, value, fatal=False):
        """
        According to the DOCTOR norm, try to convert value to the adequate type.

        Cf. http://www.umr-cnrm.fr/gmapdoc/IMG/pdf/coding-rules.pdf
        """
        t = key.split('%')[-1][0]
        try:
            if t in ('I', 'J', 'K', 'M', 'N'):
                try:
                    value = int(value)
                except ValueError:
                    raise ValueError('unable to convert variable {} from {} to int'.
                                     format(key, str(value)))
            elif t in ('L',):
                try:
                    value = bool(value)
                except ValueError:
                    raise ValueError('unable to convert variable {} from {} to bool'.
                                     format(key, str(value)))
            elif t in ('C',):
                try:
                    value = str(value)
                except ValueError:
                    raise ValueError('unable to convert variable {} from {} to str'.
                                     format(key, str(value)))
            else:
                try:
                    value = float(value)
                except ValueError:
                    raise ValueError('unable to convert variable {} from {} to float'.
                                     format(key, str(value)))
        except ValueError:
            if fatal:
                raise
        return value

    # Abstract methods that should be made available by concrete classes

    @abc.abstractmethod
    def __contains__(self, item):
        """Check if a block exists in the present namelist's set."""
        pass

    @abc.abstractmethod
    def __getitem__(self, item):
        """Retrieve the content of a namelist block.

        :return: The namelist block content
        :rtype: any subclass of collections_abc.Mapping
        """
        pass

    @abc.abstractmethod
    def __iter__(self):
        """Iterate through namelist blocks."""
        pass

    @abc.abstractmethod
    def __len__(self):
        """The number of namelist blocks."""
        pass

    @abc.abstractmethod
    def _actual_newblock(self, item):
        """Add a new block into the present namelist's set."""
        pass

    @abc.abstractmethod
    def _actual_rmblock(self, item):
        """Remove a block from the present namelist's set."""
        pass

    @abc.abstractmethod
    def _actual_mvblock(self, item, targetitem):
        """Move/Rename a block within the present namelist's set."""
        pass

    @abc.abstractmethod
    def _actual_newkey(self, block, key, value, index=None):
        """Set a new key in the present namelist's set."""
        pass

    @abc.abstractmethod
    def _actual_rmkey(self, block, key):
        """Remove a namelist keys from the present namelist's set."""
        pass

    @abc.abstractmethod
    def _actual_squeeze(self):
        """Squeeze the namelist: remove empty blocks."""
        pass

    @abc.abstractmethod
    def dumps(self, sorting=NO_SORTING):
        """
        Returns a string that represent the namelist's set (i.e. something
        readable by Fortran !)

        :param int sorting: The kind of sorting to apply within blocks
        """
        pass

    @abc.abstractmethod
    def merge(self, other):
        """Merge another namelist in the current one.

        :param AbstractNamelistAdapter other: Another namelist to merge in.
        """
        pass


class AbstractMapableNamelistAdapter(AbstractNamelistAdapter):
    """
    A generic NamelistAdapter that presumes that self._parser is itself some
    kind of Mapping object.
    """

    @staticmethod
    def _assert_mapping(obj):
        assert isinstance(obj, collections_abc.Mapping), \
            'The "{!r}" must be some kind of Mapping.'

    def __contains__(self, item):
        """Check if a block exists in the present namelist's set."""
        self._assert_mapping(self._parser)
        return item in self._parser

    def __getitem__(self, item):
        """Retrieve the content of a namelist block.

        :return: The namelist block content
        :rtype: any subclass of collections_abc.Mapping
        """
        self._assert_mapping(self._parser)
        value = self._parser[item]
        self._assert_mapping(value)
        return value

    def __iter__(self):
        """Iterate through namelist blocks."""
        self._assert_mapping(self._parser)
        return self._parser.__iter__()

    def __len__(self):
        """The number of namelist blocks."""
        self._assert_mapping(self._parser)
        return len(self._parser)


class BronxNamelistAdapter(AbstractMapableNamelistAdapter):
    """
    A NamelistAdapter that relies on the namelist parser provided by the
    :mod:`bronx` package.
    """

    def __init__(self, namelistsfile, macros=None):
        super(BronxNamelistAdapter, self).__init__(namelistsfile)
        # Delay the import of the bronx library since one may want to use another backend
        from bronx.datagrip import namelist
        actual_macros = self._all_macros(macros)
        self._parser = namelist.namparse(namelistsfile, macros=actual_macros)
        for macro, value in actual_macros.items():
            self._parser.setmacro(macro, value)

    def _actual_newblock(self, item):
        self.parser.newblock(item)

    def _actual_rmblock(self, item):
        del self.parser[item]

    def _actual_mvblock(self, item, targetitem):
        self.parser.mvblock(item, targetitem)

    def _actual_newkey(self, block, key, value, index=None):
        self[block].setvar(key, value, index=index)

    def _actual_rmkey(self, block, key):
        del self[block][key]

    def _actual_squeeze(self):
        for b in list(self.parser.keys()):
            if len(self.parser[b]) == 0:
                self._actual_rmblock(b)

    def dumps(self, sorting=NO_SORTING):
        """Returns a string that represent the namelist's set."""
        from bronx.datagrip import namelist as bnamelists
        sorting_map = dict(NO_SORTING=bnamelists.NO_SORTING,
                           FIRST_ORDER_SORTING=bnamelists.FIRST_ORDER_SORTING,
                           SECOND_ORDER_SORTING=bnamelists.SECOND_ORDER_SORTING)
        return self.parser.dumps(sorting=sorting_map.get(sorting, sorting))

    def merge(self, other):
        """Merge another namelist in the current one.

        :param AbstractNamelistAdapter other: Another namelist to merge in.
        """
        assert isinstance(other, self.__class__)
        self.parser.merge(other.parser)



# -*- coding: utf-8 -*-

"""
Utility methods widely used in the various TNT utilities.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import contextlib
import glob
import io
import logging
import os
import shutil

from bronx.fancies import loggers
from bronx.fancies.colors import termcolors
from .namadapter import BronxNamelistAdapter, NO_SORTING, FIRST_ORDER_SORTING, SECOND_ORDER_SORTING
from .config import TntRecipe

tntlog = loggers.getLogger('tntlog')
tntstacklog = loggers.getLogger('tntstacklog')


@contextlib.contextmanager
def set_verbose(verbose, filename):
    """Set or reset the verbosity level."""
    new_lformat = logging.Formatter(
        fmt=('# [%(name)s@{:s}][%(funcName)s:%(lineno)04d][%(levelname)s]: %(message)s'
             .format(os.path.basename(filename)))
    )
    old_lformat = loggers.default_console.formatter
    old_tntstack_level = tntstacklog.level
    old_tnt_level = tntlog.level
    loggers.default_console.setFormatter(new_lformat)
    tntstacklog.setLevel('INFO')
    tntlog.setLevel('INFO' if verbose else 'WARNING')
    try:
        yield
    finally:
        tntstacklog.setLevel(old_tntstack_level)
        tntlog.setLevel(old_tnt_level)
        loggers.default_console.setFormatter(old_lformat)


def process_namelist(filename, directives,
                     # options
                     sorting=NO_SORTING,
                     blocks_ref=None,
                     in_place=False,
                     outfilename=None,
                     doctor=False,
                     keep_index=False,
                     squeeze=False):
    """
    For the syntax of keys & blocks arguments, please refer to the according
    functions.

    The order of processing is that of the arguments. As movings are done first,
    check consistency.

    If **macros** is not None, it can contain the macros a.k.a. values to be
    replaced, e.g.: {'NPROC':8, 'substrA':None} will replace all NPROC values
    by 8 and will let substrA untouched.

    Other options:

    :param in_place: if True, the namelist is written back in the same file;
                     else (default), the target namelist is suffixed with '.tnt'
                     if not given as **outfilename**.
    :param outfilename: target file for out namelist
    :param sorting: Sorting option (from bronx.datagrip.namelist):
                    NO_SORTING;
                    FIRST_ORDER_SORTING => sort all keys within blocks;
                    SECOND_ORDER_SORTING => sort only within indexes or
                    attributes of the same key, within blocks.
    :param blocks_ref: if not None, defines the path for a reference namelist to
                       which the set of blocks is asserted to be equal.
    :param doctor: if True, try to convert value to DOCTOR norm according type
                   for moved keys
    :param keep_index: if True, moved keys in identical block keep the original
                       index of key in block (except a sorting is requested
                       later on.
    :param squeeze: squeeze the namelist: remove empty blocks.
    """
    if not isinstance(directives, (list, tuple)):
        directives = [directives, ]

    # The initial namelist
    initial_nam = BronxNamelistAdapter(filename, macros=directives[0].macros)

    # Target namelist file
    if not in_place:
        target_namfile = outfilename or filename + '.tnt'
    else:
        target_namfile = filename
        if outfilename is not None:
            raise ValueError("Incompatibility between arguments *outfilename* and *in_place*.")

    for d in directives:
        # process (in the appropriate order !)
        if d.new_blocks is not None:
            initial_nam.add_blocks(d.new_blocks)
        if d.blocks_to_move is not None:
            initial_nam.move_blocks(d.blocks_to_move)
        if d.keys_to_move is not None:
            initial_nam.move_keys(d.keys_to_move,
                                  doctor=doctor, keep_index=keep_index)
        if d.keys_to_remove is not None:
            initial_nam.remove_keys(d.keys_to_remove)
        if d.keys_to_set is not None:
            initial_nam.add_keys(d.keys_to_set)
        if d.blocks_to_remove is not None:
            initial_nam.remove_blocks(d.blocks_to_remove)
        if d.namdelta is not None:
            try:
                ndelta = BronxNamelistAdapter(d.namdelta, macros=d.macros)
            except ValueError:
                tntlog.error("Error while parsing the following namelist's delta:\n%s",
                             d.namdelta)
                raise
            initial_nam.merge(ndelta)

    if squeeze:
        initial_nam.squeeze()

    if blocks_ref is not None:
        cb = initial_nam.check_blocks(blocks_ref, directives[0].macros)
        if len(cb) != 0:
            tntlog.warning('Set of blocks is different from reference: ' + blocks_ref)
            tntlog.warning('diff: ' + str(cb))

    with io.open(target_namfile, 'w', encoding='ascii') as fh_namout:
        fh_namout.write(initial_nam.dumps(sorting=sorting))


def process_tnt_stack(directive, sorting=SECOND_ORDER_SORTING):
    """Apply *directive* to the current working directory.

    :param TntStackDirective directive: The tntstack directive to apply
    :param sorting: Sorting option (from bronx.datagrip.namelist):
                    NO_SORTING;
                    FIRST_ORDER_SORTING => sort all keys within blocks;
                    SECOND_ORDER_SORTING => sort only within indexes or
                    attributes of the same key, within blocks.
    """
    # Record the list of file contained in the directory
    initial_files = set()
    initial_subdirectories = set()
    for root, directories, files in os.walk('.'):
        for f in files:
            initial_files.add(os.path.normpath(os.path.join(root, f)))
        for d in directories:
            initial_subdirectories.add(os.path.normpath(os.path.join(root, d)))

    # Process the todolist
    for todo in directive.todolist:
        action = todo['action']

        if action == 'tnt':
            for nam in todo['namelist']:
                for realnam in glob.glob(nam):
                    tntstacklog.info("Namelist '%s': applying the following directives: %s",
                                     realnam, ",".join(todo['directive']))
                    process_namelist(realnam,
                                     [directive.directives[name] for name in todo['directive']],
                                     in_place=True, sorting=sorting)
                    initial_files.discard(os.path.normpath(realnam))

        elif action == 'create':
            if 'external' in todo:
                tntstacklog.info("Creating namelist '%s' from external file '%s'", todo['target'], todo['external'])
                shutil.copy(todo['external'], todo['target'])
            elif 'copy' in todo:
                tntstacklog.info("Creating namelist '%s' from file '%s'", todo['target'], todo['copy'])
                shutil.copy(todo['copy'], todo['target'])
            else:
                tntstacklog.info("Creating namelist '%s' from namelist '%s' by applying the following directives: %s",
                                 todo['target'], todo['namelist'], ",".join(todo['directive']))
                process_namelist(todo['namelist'],
                                 [directive.directives[name] for name in todo['directive']],
                                 outfilename=todo['target'], sorting=sorting)
            initial_files.discard(os.path.normpath(todo['target']))

        elif action in ('delete', 'touch'):
            for nam in todo['namelist']:
                for realnam in glob.glob(nam):
                    if action == 'delete':
                        tntstacklog.info("Deleting namelist '%s'", realnam)
                        os.unlink(realnam)
                    elif action == 'touch':
                        tntstacklog.info("Marking file '%s' as touched'", realnam)
                    initial_files.discard(os.path.normpath(realnam))

        elif action == 'link':
            tntstacklog.info("Linking '%s' -> '%s'", todo['target'], todo['namelist'])
            os.symlink(todo['namelist'], todo['target'])
            initial_files.discard(os.path.normpath(todo['namelist']))

        elif action == 'move':
            tntstacklog.info("Moving '%s' to '%s'", todo['namelist'], todo['target'])
            shutil.move(todo['namelist'], todo['target'])
            initial_files.discard(os.path.normpath(todo['namelist']))
            initial_files.discard(os.path.normpath(todo['target']))

        elif action == 'clean_untouched':
            for f in initial_files:
                tntstacklog.info("Deleting file '%s'", f)
                os.unlink(f)
            for d in [d for d in initial_subdirectories if not os.listdir(d)]:
                # Remove empty directories
                tntstacklog.info("Deleting empty directory '%s'", d)
                os.rmdir(d)


def namelist_read_and_sort(namfile):
    """Read a namelist and return it as a sorted string."""
    try:
        namp = BronxNamelistAdapter(namfile)
    except (ValueError, IOError):
        tntlog.error("Something went wrong will reading: %s", namfile)
        raise
    if not len(namp):
        raise ValueError('Nothing to read in "{:s}": Is it a namelist ?'.format(namfile))
    return namp.dumps(sorting=FIRST_ORDER_SORTING)


def _check_diffline(line, expected):
    return line and len(line) >= 2 and line[0] == expected and line[1] == ' '


def _color_diffline(prevline, line=''):
    """Colorise an individual line of a difflib output."""
    if _check_diffline(prevline, '-') and _check_diffline(line, '?'):
        return termcolors.error(prevline)
    elif _check_diffline(prevline, '+') and _check_diffline(line, '?'):
        return termcolors.error(prevline)
    elif _check_diffline(prevline, '-'):
        return termcolors.critical(prevline)
    elif _check_diffline(prevline, '+'):
        return termcolors.okgreen(prevline)
    elif _check_diffline(prevline, '?'):
        return termcolors.error(prevline)
    else:
        return prevline


def colorise_diff(lines):
    """Colorise the lines of any kind of difflib outputs."""
    newdiff = list()
    prevline = None
    for l in lines:
        if prevline:
            newdiff.append(_color_diffline(prevline, l))
        prevline = l
    newdiff.append(_color_diffline(prevline))
    return newdiff


def compose_namelist(recipe_filename,
                     sourcenam_directory=None,
                     suffix='.nam',
                     sorting=NO_SORTING,
                     squeeze=False,
                     fhoutput=None):
    """
    Compose a namelist from a **recipe_filename**. For the syntax of recipe,
    see template recipe.

    :param recipe_filename: The name of the recipe/directive YAML file
    :param sourcenam_directory: path to directory in which to look for source
                                namelists
    :param suffix: suffix to add to generated namelist:
                   output name is ./recipe_basename[suffix]
    :param sorting: Sorting option (from bronx.datagrip.namelist):
                    NO_SORTING;
                    FIRST_ORDER_SORTING => sort all keys within blocks;
                    SECOND_ORDER_SORTING => sort only within indexes or
                    attributes of the same key, within blocks.
    :param squeeze: squeeze the namelist: remove empty blocks.
    :param fhoutput: a file object where the result is written (if None, a
                     new file named `basename(recipe_filename)` is created).
    """
    # read
    recipe = TntRecipe(recipe_filename, sourcenam_directory=sourcenam_directory)
    # merge
    nam = recipe.ingredients[0]
    for ingredient in recipe.ingredients[1:]:
        nam.merge(ingredient)
    if squeeze:
        nam.squeeze()
    # write
    if fhoutput is None:
        namelistname = os.path.basename(recipe_filename.replace('.yaml', suffix))
        with io.open(namelistname, 'w', encoding='ascii') as fh_namout:
            fh_namout.write(nam.dumps(sorting=sorting))
    else:
        fhoutput.write(nam.dumps(sorting=sorting))



from __future__ import print_function, absolute_import, unicode_literals, division

import os
import unittest

from thenamelisttool.namadapter import BronxNamelistAdapter, NO_SORTING

tpl_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        '../src/thenamelisttool/templates')
tpl_path = os.path.normpath(tpl_path)


TEST_FINALREF = """ &NAM_FILENAMES
   HPGDFILE='PGDFILE',
 /
 &NAM_IO
   CSURF_FILETYPE='LFI   ',
   GLURP='PREP2',
   CPGDFILE='PGD2',
   CINIFILEBIS='INIT_SURF',
   NPRINT=1,
 /
 &NAM_ISBAN
   LGLACIER=.FALSE.,
 /
 &NAM_PREP_SEAFLUX
 /
 &NAM_PREP_SURF_ATM
   CFILETYPE='LFI   ',
   CFILE='PREP1',
   CFILEPGDTYPE='LFI   ',
   CFILEPGD=55.6,
 /
 &NAM_PREP_WATFLUX
   LWAT_SBL=.FALSE.,
 /
 &NAM_TOTO
 /
 &NAM_WRITE_SURF_ATM
 /
"""


class TestTntNamAdapter(unittest.TestCase):

    def test_bronx_adapter(self):
        nampath = os.path.join(tpl_path, 'namelist_prep_template')
        nadapt = BronxNamelistAdapter(nampath)
        nadapt.remove_blocks(['nam_PREP_ISBA', ])
        nadapt.add_blocks(['NAM_TOTO'])
        nadapt.move_blocks({'NAM_FILE_NAMES': 'NAM_FILENAMES',
                            'NAM_IO_OFFLINE': 'NAM_IO'})
        nadapt.move_keys({('NAM_FILENAMES', 'CINIFILE'): ('NAM_IO', 'CINIFILEBIS')})
        nadapt.move_keys({('NAM_IO', 'LPRINT'): ('NAM_IO', 'NPRINT')}, doctor=True)
        nadapt.move_keys({('NAM_IO', 'CPREPFILE'): ('NAM_IO', 'GLURP')}, keep_index=True)
        with self.assertRaises(KeyError):
            nadapt.add_keys({('NAMTOTO', 'TRUC'): 1})
        nadapt.add_keys({('NAM_ISBAN', 'LGLACIER'): False,
                         ('NAM_PREP_SURF_ATM', 'CFILEPGD'): 55.6, })
        nadapt.remove_keys([('NAM_WRITE_SURF_ATM', 'LNOWRITE_TEXFILE'),
                            ('NAM_PREP_SEAFLUX', 'LSEA_SBL'), ])
        self.assertEqual(nadapt.dumps(sorting=NO_SORTING), TEST_FINALREF)
        self.assertSetEqual(set(['NAM_IO', 'NAM_IO_OFFLINE', 'NAM_FILE_NAMES', 'NAM_FILENAMES', 'NAM_TOTO', 'NAM_PREP_ISBA']),
                            nadapt.check_blocks(nampath))
        nadapt.merge(BronxNamelistAdapter('&NAM_PREP_WATFLUX LWAT_SBL=T, /'))
        self.assertIs(nadapt['NAM_PREP_WATFLUX']['LWAT_SBL'], True)
        self.assertIn('NAM_TOTO', nadapt)
        nadapt.squeeze()
        self.assertNotIn('NAM_TOTO', nadapt)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

"""
A bunch of classes and functions that deal with TNT configuration files and
directives.
"""

from __future__ import absolute_import, division, print_function, unicode_literals
import six

import collections
import io
import os
import pprint
import re
import string
import sys

from bronx.compat.moves import collections_abc
from bronx.fancies import loggers
from bronx.syntax.decorators import secure_getattr
from .namadapter import BronxNamelistAdapter

tntlog = loggers.getLogger('tntlog')

# Detect TNT's configuration files directory
TPL_DIRECTORY = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             'templates')


# EXCEPTIONS
#

class TntDirectiveError(Exception):
    """Any exception raised by a TNT configuration class should derive form this one."""
    pass


class TntDirectiveUnkownError(TntDirectiveError):
    """The TNT directive is unknown."""

    def __init__(self, name):
        errmsg = 'The "{:s}" directive is not allowed with TNT.'.format(name)
        super(TntDirectiveUnkownError, self).__init__(errmsg)


class TntDirectiveValueError(TntDirectiveError):
    """An inappropriate value was given as a TNT directive."""

    def __init__(self, name, value):
        errmsg = '"{!s}" is not an appropriate value for the "{:s}" directive.'.format(value, name)
        super(TntDirectiveValueError, self).__init__(errmsg)


class TntStackDirectiveError(TntDirectiveError):
    """Syntax error in the TNT stack directive file."""
    pass


# TNT directives part
#

class TntDirective(object):
    """Class that holds all the necessary informations about TNT directives.

    It is in charge of checking the correctness of all the provided attributes.
    """

    _ALLOWED_DIRECTIVES = set(('keys_to_remove', 'keys_to_set', 'keys_to_move',
                               'blocks_to_move', 'blocks_to_remove', 'new_blocks',
                               'macros', 'namdelta'))

    def _check_keytuple(self, val, theexc, unique=False):
        """Verify if *val* is some kind of a (block, key) tuple description."""
        keylist = list()
        # Reformat dictionaries, iterables, ...
        if isinstance(val, collections_abc.Mapping):
            for b, v in val.items():
                if isinstance(v, collections_abc.Iterable) and not isinstance(v, six.string_types):
                    keylist.extend([(b, kk) for kk in v])
                else:
                    keylist.append((b, v), )
        elif isinstance(val, collections_abc.Iterable):
            if all([isinstance(v, collections_abc.Iterable) and not isinstance(v, six.string_types)
                    for v in val]):
                keylist.extend(val)
            else:
                keylist.append(tuple(val))
        else:
            raise theexc
        # Final check...
        if not all([len(k) == 2 and
                    isinstance(k[0], six.string_types) and
                    isinstance(k[1], six.string_types)
                    for k in keylist]):
            raise theexc
        # Unique element wanted ?
        if unique:
            if len(keylist) > 1:
                raise theexc
            return keylist[0]
        else:
            return set(keylist)

    def _process_keys_to_remove(self, val):
        myexc = TntDirectiveValueError('keys_to_remove', val)
        return self._check_keytuple(val, myexc)

    def _process_keys_to_set(self, val, theexc=None):
        kdict = dict()
        myexc = theexc or TntDirectiveValueError('keys_to_set', val)
        if isinstance(val, collections_abc.Mapping):
            for k, v in val.items():
                if isinstance(v, collections_abc.Mapping):
                    for kk, vv in v.items():
                        kdict[self._check_keytuple((k, kk), myexc, unique=True)] = vv
                elif isinstance(k, collections_abc.Iterable):
                    kdict[self._check_keytuple(k, myexc, unique=True)] = v
                else:
                    raise myexc
        else:
            raise myexc
        return kdict

    def _process_keys_to_move(self, val):
        kdict = dict()
        myexc = TntDirectiveValueError('keys_to_move', val)
        keystructure = self._process_keys_to_set(val, theexc=myexc)
        for k, v in keystructure.items():
            kdict[k] = self._check_keytuple(v, myexc, unique=True)
        return kdict

    def _process_blocks_to_move(self, val):
        if not (isinstance(val, collections_abc.Mapping) and
                all([isinstance(k, six.string_types) and isinstance(v, six.string_types)
                     for k, v in val.items()])):
            raise TntDirectiveValueError('blocks_to_move', val)
        return {k: v for k, v in val.items()}

    def _process_set_of_blocks(self, val, realname):
        if (isinstance(val, collections_abc.Iterable) and
                not isinstance(val, six.string_types) and
                all([isinstance(v, six.string_types) for v in val])):
            return set(val)
        elif isinstance(val, six.string_types):
            return set([val, ])
        else:
            raise TntDirectiveValueError(realname, val)

    def _process_blocks_to_remove(self, val):
        return self._process_set_of_blocks(val, 'blocks_to_remove')

    def _process_new_blocks(self, val):
        return self._process_set_of_blocks(val, 'new_blocks')

    def _process_macros(self, val):
        if not (isinstance(val, collections_abc.Mapping) and
                all([isinstance(k, six.string_types) for k in val.keys()])):
            raise TntDirectiveValueError('macros', val)
        return {k: v for k, v in val.items()}

    def _process_namdelta(self, val):
        if not (isinstance(val, six.string_types)):
            raise TntDirectiveValueError('namdelta', val)
        return val

    def __init__(self, **kwargs):
        self._internals = dict()
        # Is the directive allowed ?
        for k, v in kwargs.items():
            if k not in self._ALLOWED_DIRECTIVES:
                raise TntDirectiveUnkownError(k)
            self._internals[k] = getattr(self, '_process_{:s}'.format(k))(v)

    @secure_getattr
    def __getattr__(self, item):
        if item not in self._ALLOWED_DIRECTIVES:
            raise TntDirectiveUnkownError(item)
        else:
            return self._internals.get(item, None)


def read_directives(filename):
    """
    Read TNT directives in an external file (**filename**).

    For a template of directives, call function *write_directives_template()*.
    """
    if os.path.splitext(filename)[1] in ('.yaml', '.yml'):
        import yaml
        with io.open(filename, 'r') as yamlfh:
            return TntDirective(**yaml.load(yamlfh, Loader=yaml.SafeLoader))
    else:
        prev_bytecode_flag = sys.dont_write_bytecode
        try:
            sys.dont_write_bytecode = True
            if sys.version_info.major == 3 and sys.version_info.minor >= 4:
                from importlib.machinery import SourceFileLoader  # @UnresolvedImport
                import importlib.util as imputil  # @UnresolvedImport
                spec = imputil.spec_from_loader('thenamelisttool.tnt_raw_configuration_module',
                                                SourceFileLoader('thenamelisttool.tnt_raw_configuration_module',
                                                                 os.path.abspath(filename)))
                m = imputil.module_from_spec(spec)
                spec.loader.exec_module(m)
            else:
                import imp  # @UnresolvedImport
                m = imp.load_source(filename, os.path.abspath(filename))
        finally:
            sys.dont_write_bytecode = prev_bytecode_flag
        return TntDirective(**{k: v for k, v in m.__dict__.items() if not k.startswith('_')})


# TNTstack directives part
#

class TntStackDirective(object):
    """Class that holds all the necessary informations about TNTstack directives.

    :param str basedir: The name of the directory where the TNTstack reqests lies
    :param list[dict] todolist: The list of actions to be performed
    :param dict[dict] directives: A list of TNT directives
    """

    def __init__(self, basedir, todolist, directives=None):
        self._basedir = basedir
        self._directives = dict()
        self._todolist = list()
        self._directives_init(directives or dict())
        self._todolist_init(todolist)

    def _directives_init(self, directives):
        """
        Read the *directives* attribute and create :class:`TntDirective` objects
        from that.
        """
        if not (isinstance(directives, collections_abc.Mapping) and
                all([isinstance(v, collections_abc.Mapping) for v in directives.values()])):
            raise TntStackDirectiveError('The directives argument must be a mapping of mappings')
        for k, v in directives.items():
            if 'external' in v:
                newdir = read_directives(os.path.join(self._basedir, v['external']))
            else:
                newdir = TntDirective(**v)
            self._directives[k] = newdir

    def _checkdict(self, action, values, attr, str_or_list=False):
        """Check that the *values* dictionary has a valid *attr* item.

        :param str action: The name of the current action (used to print meaningful
                           error messages).
        :param bool str_or_list: It *True*, the *attr* item can be either a string
                                 or a list of strings. (otherwise, it has to be a string).
        """
        stuff = values.get(attr, None)
        if stuff is None:
            tntlog.error("Error while processing todo list item:\n%s", values)
            raise TntStackDirectiveError('The "{:s}" entry is mandatory with the "{:s}" action'
                                         .format(attr, action))
        if isinstance(stuff, six.string_types) or not isinstance(stuff, collections_abc.Iterable):
            stuff = [stuff, ]
        else:
            stuff = [v for v in stuff]
        if str_or_list:
            return stuff
        else:
            if len(stuff) != 1:
                tntlog.error("Error while processing todo list item:\n%s", values)
                raise TntStackDirectiveError('The "{:s}" entry requires a unique element ("{:s}" action).'
                                             .format(attr, action))
            return stuff[0]

    def _todolist_init(self, todolist):
        """Read the *todolist* attribute and check that everything is ok."""
        if not (isinstance(todolist, collections_abc.Iterable) and
                all([isinstance(v, collections_abc.Mapping) for v in todolist])):
            raise TntStackDirectiveError('The todolist argument must be an iterable of mappings')
        for todo in todolist:
            action = todo.get('action', None)
            action_d = dict(action=action)

            if action == 'tnt':
                action_d['namelist'] = self._checkdict(action, todo, 'namelist', str_or_list=True)
                action_d['directive'] = self._checkdict(action, todo, 'directive', str_or_list=True)

            elif action == 'create':
                action_d['target'] = self._checkdict(action, todo, 'target')
                if 'external' in todo:
                    action_d['external'] = os.path.join(self._basedir,
                                                        self._checkdict(action, todo, 'external'))
                    if not os.path.isfile(action_d['external']):
                        raise TntStackDirectiveError('The "{:s}"  does not exists.'
                                                     .format(action_d['external']))
                elif 'copy' in todo:
                    action_d['copy'] = self._checkdict(action, todo, 'copy')
                else:
                    action_d['namelist'] = self._checkdict(action, todo, 'namelist')
                    action_d['directive'] = self._checkdict(action, todo, 'directive', str_or_list=True)

            elif action in ('delete', 'touch'):
                action_d['namelist'] = self._checkdict(action, todo, 'namelist', str_or_list=True)

            elif action in ('link', 'move'):
                action_d['target'] = self._checkdict(action, todo, 'target')
                action_d['namelist'] = self._checkdict(action, todo, 'namelist')

            elif action == 'clean_untouched':
                pass

            else:
                raise TntStackDirectiveError('Unknown action "{!s}" requested in the todolist'.format(action))

            # Check the directive entry against existing directives
            if 'directive' in action_d:
                tocheck = ([action_d['directive'], ]
                           if isinstance(action_d['directive'], six.string_types)
                           else action_d['directive'])
                for a_dir in tocheck:
                    if a_dir not in self.directives:
                        raise TntStackDirectiveError('The "{:s}" directive is not defined.'.format(a_dir))
            self._todolist.append(action_d)

    @property
    def directives(self):
        """The dictionary of TNT directives (as :class:`TntDirective` objects)."""
        return self._directives

    @property
    def todolist(self):
        """The todo's list (as a list of dictionaries)."""
        return self._todolist


class TntRecipeSyntaxError(ValueError):
    """Raised when a syntax error is detected in the recipe file."""
    pass


class TntRecipe(object):
    """
    A YAML Recipe reader, that collects namelists and possibly filter them,
    in sight of finally merging them.
    """

    _ingredient_name_re = re.compile(r'(?P<nam>.+?)(?:/(?P<filter>(?:-|\+)))?$')
    _ingredient_item_re = re.compile(r'(?P<block>[^/]+)/(?P<filter>(?:-|\+))$')

    def __init__(self, recipe_filename, sourcenam_directory=None):
        """
        :param recipe_filename: filepath to the YAML recipe
        :param sourcenam_directory: an optional external directory in which to
            pick the ingredient namelists
        """
        self.sourcenam_directory = sourcenam_directory
        self._load_recipe(recipe_filename)

    def _throw_syntax_err(self, entry, wholeentry, msg):
        """Deal with a syntax error."""
        tntlog.critical("Syntax error in the '%s' entry: %s.", entry, msg)
        if wholeentry is not None:
            tntlog.critical("The '%s' entry content is:\n%s.", entry,
                            pprint.pformat(wholeentry, indent=2))
        raise TntRecipeSyntaxError('Syntax error in the {:s} entry of the Recipe file.'
                                   .format(entry))

    def _read_init_final_elements(self, what, ingredient):
        """Read '__initial__' or '__final__' step **ingredient**."""
        if ingredient is not None:
            if isinstance(ingredient, six.string_types):
                # external namelist
                if self.sourcenam_directory:
                    ingredient = os.path.join(self.sourcenam_directory, ingredient)
                nam = BronxNamelistAdapter(ingredient, macros=self.macros)
            elif isinstance(ingredient, dict):
                # internal dict/yaml namelist
                nam = BronxNamelistAdapter(six.StringIO(), macros=self.macros)
                nam.add_blocks(list(ingredient.keys()))
                keys_to_add = {}
                for b, kv in ingredient.items():
                    if kv is not None:
                        if isinstance(kv, dict):
                            for k, v in ingredient[b].items():
                                keys_to_add[(b, k)] = v
                        else:
                            self._throw_syntax_err(what, ingredient,
                                                   "'{!s}': should be 'null' or a dictionary"
                                                   .format(b))
                nam.add_keys(keys_to_add)
            else:
                self._throw_syntax_err(what, ingredient,
                                       "Should be 'null', a string or a dictionary")
        else:
            nam = BronxNamelistAdapter(six.StringIO(), macros=self.macros)
        return nam

    def _process_ingredient(self, input_nam, blocks):
        """Preprocess ingredient: read according namelist and filter it."""
        # Check input_nam
        input_nam_m = self._ingredient_name_re.match(input_nam)
        if not input_nam_m:
            self._throw_syntax_err(input_nam, None, "Invalid ingredient name.")
        # read namelist
        input_nam_filename = input_nam_m.group('nam')
        if self.sourcenam_directory:
            input_nam_filename = os.path.join(self.sourcenam_directory,
                                              input_nam_filename)
        ingredient = BronxNamelistAdapter(input_nam_filename,
                                          macros=self.macros)
        # prepare filtering elements
        blocks_filter = collections.defaultdict(list)
        keys_filter = collections.defaultdict(dict)
        # Process the various blocks provided by the user
        if isinstance(blocks, (list, tuple, set)):
            for b in blocks:
                if isinstance(b, six.string_types):
                    blocks_filter[input_nam_m.group('filter')].append(b)
                if isinstance(b, dict) and len(b) == 1:
                    bname, blist = list(b.items())[0]
                    bname_m = self._ingredient_item_re.match(bname)
                    if bname_m:
                        if not isinstance(blist, (list, tuple, set)):
                            self._throw_syntax_err(input_nam, blocks,
                                                   "'{:s}' is not a list".format(blist))
                        keys_filter[bname_m.group('filter')][bname_m.group('block')] = set(blist)
                    else:
                        self._throw_syntax_err(input_nam, blocks,
                                               "invalid namelist block definition: '{!s}'"
                                               .format(bname))
            if input_nam_m.group('filter') == '+':  # Include only case
                authorized_blocks = set(blocks_filter['+'])
                for a_keys_filter in keys_filter.values():
                    authorized_blocks.update(a_keys_filter.keys())
                blocks_filter['-'] = list(set(ingredient) - authorized_blocks)
        elif isinstance(blocks, six.string_types) and blocks == '__all__':
            pass  # Ok, includes everything
        else:
            self._throw_syntax_err(input_nam, blocks, 'invalid ingredient description')
        # then filter
        # blocks that are not requested
        ingredient.remove_blocks(blocks_filter['-'])
        # keys that are excluded
        for b, keys in keys_filter['-'].items():
            ingredient.remove_keys([(b, k) for k in keys])
        # keys that are not requested
        for b, keys in keys_filter['+'].items():
            to_remove = [(b, k) for k in ingredient[b] if k not in keys]
            ingredient.remove_keys(to_remove)
        return ingredient

    def _load_recipe(self, recipe_filename):
        """Read YAML file and preprocess ingredients."""
        from bronx.datagrip.misc import load_ordered_yaml
        # load yaml
        recipe = load_ordered_yaml(recipe_filename)
        if not isinstance(recipe, collections.OrderedDict):
            raise TntRecipeSyntaxError('The recipe must be a dictionary.')
        # specific cases: initialization, finalization, macros
        self.macros = recipe.pop('__macros__', {})
        initial = self._read_init_final_elements('__initial__',
                                                 recipe.pop('__initial__', None))
        final = self._read_init_final_elements('__final__',
                                               recipe.pop('__final__', None))
        self.ingredients = []
        # convert each ingredient into a namelist to be merged
        for input_nam, blocks in recipe.items():
            self.ingredients.append(self._process_ingredient(input_nam, blocks))
        self.ingredients.insert(0, initial)
        self.ingredients.append(final)


# Utility function that deals with template files
#

def get_template(tplname, encoding=None):
    """Retrieve a template file in the dedicated directory."""
    tplfile = os.path.join(TPL_DIRECTORY, tplname)
    with io.open(tplfile, 'r', encoding=encoding) as fhtpl:
        tpl = string.Template(fhtpl.read())
    return tpl


def write_directives_template(out=sys.stdout, tplname='tnt-directive.tpl.py'):
    """Write out a directives template."""
    if isinstance(out, six.string_types):
        outfh = io.open(out, 'w')
    else:
        outfh = out
    outtpl = os.path.join(TPL_DIRECTORY, tplname)
    try:
        with io.open(outtpl, 'r') as tplfh:
            for line in tplfh:
                outfh.write(line)
    finally:
        if isinstance(out, six.string_types):
            outfh.close()



# -*- coding: utf-8 -*-

"""
TNT - The Namelist Tool - Compose: a namelist composer merging parts of others.
"""

from __future__ import print_function, absolute_import, unicode_literals, division

import argparse
import os

import thenamelisttool as tnt


def main():
    """Start the tntcompose CLI."""
    _tmpl = 'tmpl_compose-recipe.tnt'

    parser = argparse.ArgumentParser(description='TNT - The Namelist Tool - Compose: ' +
                                     'a namelist composer merging parts of others.',
                                     epilog='End of help for: %(prog)s')
    parser.add_argument('recipes',
                        type=str,
                        nargs='*',
                        help='YAML recipe file(s) of namelists to be composed.')
    parser.add_argument('-R',
                        dest='generate_recipe_template',
                        action='store_true',
                        help="generates a recipe template '{}.yaml'.".format(_tmpl))
    parser.add_argument('-d',
                        dest='sourcenam_directory',
                        default=None,
                        help="specify the directory in which to find the source namelists mentioned in recipes")
    parser.add_argument('-x',
                        dest='suffix',
                        default='.nam',
                        help='suffix to be used in output namelist filename, in replacement of .yaml')
    parser.add_argument('--squeeze',
                        dest='squeeze',
                        action='store_true',
                        help='squeeze the namelist: remove empty blocks.',
                        default=False)
    sorting = parser.add_mutually_exclusive_group()
    sorting.add_argument('-S',
                         action='store_true',
                         dest='firstorder_sorting',
                         help='First order sorting: sort all keys within blocks.',
                         default=False)
    sorting.add_argument('-s',
                         action='store_true',
                         dest='secondorder_sorting',
                         help='Second order sorting: sort only within indexes \
                               or attributes of the same key within blocks.',
                         default=False)
    parser.add_argument('-v',
                        action='store_true',
                        dest='verbose',
                        help='verbose mode.',
                        default=False)
    args = parser.parse_args()

    if args.generate_recipe_template:
        tnt.config.write_directives_template(_tmpl + '.yaml',
                                             tplname='tntcompose-recipe.tpl.yaml')
        print("Template of directives written in: " +
              os.path.abspath(_tmpl + '.yaml'))
    else:
        assert len(args.recipes) > 0, "no namelists provided to process."
        if args.firstorder_sorting:
            sorting = tnt.namadapter.FIRST_ORDER_SORTING
        elif args.secondorder_sorting:
            sorting = tnt.namadapter.SECOND_ORDER_SORTING
        else:
            sorting = tnt.namadapter.NO_SORTING
        for recipe in args.recipes:
            with tnt.util.set_verbose(args.verbose, recipe):
                tnt.util.compose_namelist(recipe,
                                          sourcenam_directory=args.sourcenam_directory,
                                          suffix=args.suffix,
                                          sorting=sorting,
                                          squeeze=args.squeeze)



# -*- coding: utf-8 -*-

"""
TNTdiff - The Namelist Tool: a namelist comparator.

Compares two namelists and produce the following outputs:

* (default)   a TNT directives file to go from one (before/-b) to the other
              (after/-a);
* (-H option) a HTML file that displays the differences in a table
* (-V or -v)  a summary of the differences on the standard output
* (-e option) a visualisation of the differences in an external tool

Beware that TNTdiff (purposely) DO NOT take into account:

* Differences in the order of appearance of namelists blocks or keys
* Differences in the formatting of namelist values (e.g 1.0 and 1.0000 are
  considered the same)

With the -H, -V, -v or -e options, the displayed namelists DO NOT necessarily
correspond to the original files since, prior to be displayed, blocks/keys are
ordered alphabetically and values are formatted in a "standard" way.

"""

from __future__ import print_function, absolute_import, unicode_literals, division

import argparse
import difflib
import io
import os
import six
import subprocess
import tempfile

from bronx.stdtypes.tracking import Tracker, MappingTracker
import thenamelisttool as tnt

_outfilename = 'tntdiff.out'


def _string_encode(valueslist):
    return [("'{:s}'".format(v) if isinstance(v, six.string_types)
             else "{}".format(v))
            for v in valueslist]


def visualdiff(before_filename, after_filename, bw=False):
    """Print a nicely formated text representation of the differences."""
    namtxtB = tnt.util.namelist_read_and_sort(before_filename).split('\n')
    namtxtA = tnt.util.namelist_read_and_sort(after_filename).split('\n')

    diff = difflib.ndiff(namtxtB, namtxtA)
    if not bw and diff:
        diff = tnt.util.colorise_diff(diff)

    print('\n'.join(diff))


def htmldiff_view(before_filename, after_filename, outfilename):
    """Create an HTML representation of the differences and open a web browser."""
    namtxtB = tnt.util.namelist_read_and_sort(before_filename).split('\n')
    namtxtA = tnt.util.namelist_read_and_sort(after_filename).split('\n')

    htmldiff = difflib.HtmlDiff()
    htmlout = htmldiff.make_file(namtxtB, namtxtA, before_filename, after_filename)
    with io.open(outfilename, "w") as fh_ht:
        fh_ht.writelines(htmlout)
    import webbrowser
    webbrowser.open(outfilename)


def extdiff(before_filename, after_filename, tool):
    """Visualise the differences with an external tool such as ``vim`` or ``meld``."""
    with tempfile.NamedTemporaryFile(mode='w', prefix='tntdiff_BEFORE.', delete=True) as fhB:
        with tempfile.NamedTemporaryFile(mode='w', prefix='tntdiff_AFTER.', delete=True) as fhA:
            fhB.write(tnt.util.namelist_read_and_sort(before_filename))
            fhA.write(tnt.util.namelist_read_and_sort(after_filename))
            fhB.flush()
            fhA.flush()
            if tool == 'meld':
                subprocess.check_call(['meld', '--diff', fhB.name, fhA.name])
            elif tool == 'vim':
                subprocess.check_call(['vim', '-d', fhB.name, fhA.name])
            else:
                raise ValueError("Unknown diff tool.")


def actual_main(before_filename, after_filename, outfilename):
    """
    Compare two namelists and return directives to go from one (before) to the
    other (after).

    :param outfilename: output file in which to store directives (.py).
                        Or None if not required.
    """
    # Read namelists
    before_namelist = tnt.namadapter.BronxNamelistAdapter(before_filename)
    after_namelist = tnt.namadapter.BronxNamelistAdapter(after_filename)

    blocks_diff = Tracker(before=before_namelist.keys(), after=after_namelist.keys())
    keys_diff = MappingTracker(before={(b, k): v for b, bl in before_namelist.items() for k, v in bl.items()},
                               after={(b, k): v for b, bl in after_namelist.items() for k, v in bl.items()},)

    # Compare:
    # 7. macros
    # 6. blocks to remove
    blocks_to_remove = blocks_diff.deleted

    # 5. Keys to be set with a value (new or modified).
    keys_to_set = {(b, k): after_namelist[b][k] for b, k in keys_diff.created}

    # 4. Keys to be removed.
    keys_to_remove = set([(b, k) for b, k in keys_diff.deleted if b in after_namelist])

    # 3. Keys to be moved. No way to discriminate from set/remove => treated this way
    # 2. Blocks to be moved. No way to discriminate from new/remove => treated this way
    # 1. Blocks to be added.
    new_blocks = blocks_diff.created

    # 0. Modified values
    modified_values = {}
    for b, k in keys_diff.updated:
        keys_to_set[(b, k)] = after_namelist[b][k]
        modified_values[(b, k)] = (getattr(before_namelist[b], 'dumps_values', str)(k),
                                   getattr(after_namelist[b], 'dumps_values', str)(k))

    # Prepare the output directives
    tplsep = ',\n' + ' ' * 4

    outtpl = tnt.config.get_template('tnt-diff-outputdir.tpl', encoding='utf_8')
    outstr = outtpl.substitute(dict(
        TPL_NEWBLOCKS=tplsep.join(_string_encode(new_blocks)),
        TPL_RMKEYS=tplsep.join(_string_encode(keys_to_remove)),
        TPL_NEWKEYS=tplsep.join(['{}: {}'.format(k, v)
                                 for k, v in zip(* map(_string_encode,
                                                       zip(* sorted(keys_to_set.items()))))]),
        TPL_MODIFIED=('#None' if not modified_values
                      else '\n'.join(["# {}: {} => {}".format(k, *v)
                                      for k, v in sorted(modified_values.items())])),
        TPL_RMBLOCKS=tplsep.join(_string_encode(blocks_to_remove)),
    ))

    # Write to file
    if outfilename is not None:
        with io.open(outfilename, 'w', encoding='utf_8') as outfh:
            outfh.write(outstr)


def main():
    """Start the tntdiff CLI."""
    program_desc = '%(prog)s -- ' + __import__('__main__').__doc__.lstrip('\n')
    parser = argparse.ArgumentParser(description=program_desc, epilog='End of help for: %(prog)s',
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-b', '--before',
                        required=True,
                        help="source namelist.")
    parser.add_argument('-a', '--after',
                        required=True,
                        help="target namelist.")
    parser.add_argument('-o',
                        default=_outfilename,
                        dest='outputfilename',
                        help="output filename (without any extension). Defaults to %(default)s.")
    visual = parser.add_mutually_exclusive_group()
    visual.add_argument('-H',
                        action='store_true',
                        dest='html',
                        help="Create a HTML file and display it in a webbrowser.")
    visual.add_argument('-V',
                        action='store_true',
                        dest='visual',
                        help="Visualise the diff result on the standard output.")
    visual.add_argument('-v',
                        action='store_true',
                        dest='visualbw',
                        help="Visualise the diff result on the standard output (in black & white).")
    visual.add_argument('-e',
                        choices=('meld', 'vim'),
                        dest='external',
                        help="Use an external tool to compute and display the diff.")
    args = parser.parse_args()
    if args.html:
        print("HTML diff file written in: " + os.path.abspath(args.outputfilename + '.html'))
        htmldiff_view(args.before, args.after, args.outputfilename + '.html')
    elif args.visual or args.visualbw:
        visualdiff(args.before, args.after, bw=args.visualbw)
    elif args.external:
        extdiff(args.before, args.after, args.external)
    else:
        print("Diff directives written in: " + os.path.abspath(args.outputfilename + '.py'))
        actual_main(args.before, args.after, args.outputfilename + '.py')
